@@ -20,7 +20,8 @@ else:
 import bisect
 from pyasn1.type import univ
 from pyasn1.codec.ber import encoder, decoder
-from pyasn1.compat.octets import octs2str
+from pyasn1.compat.octets import octs2str, str2octs
+from pyasn1.error import PyAsn1Error
 from pysnmp.entity import engine, config
 from pysnmp.entity.rfc3413 import cmdrsp, context
 from pysnmp.carrier.asynsock.dgram import udp
@@ -35,6 +36,7 @@ except ImportError:
 from pysnmp.carrier.asynsock.dispatch import AsynsockDispatcher
 from pysnmp.smi import exval, indices
 from pysnmp.proto import rfc1902, rfc1905, api
+from pysnmp import error
 from pysnmp import debug
 
 # Process command-line options
@@ -54,7 +56,6 @@ agentUDPv4Endpoints = []
 agentUDPv6Endpoints = []
 agentUNIXEndpoints = []
 deviceDirs = set()
-deviceExt = os.path.extsep + 'snmpwalk'
 
 authProtocols = {
   'MD5': config.usmHMACMD5AuthProtocol,
@@ -227,7 +228,7 @@ class MvcParser(DumpParser):
     ext = os.path.extsep + 'MVC'  # just an alias
 
 class SnmprecParser:
-    ext = os.path.extsep + 'snmprec'
+    ext = os.path.extsep + 'snmprec' 
     tagMap = {}
     for t in ( rfc1902.Gauge32,
                rfc1902.Integer32,
@@ -349,10 +350,14 @@ class DeviceFile:
                     oid, tag, val = textParser.parse(line)
                 except Exception:
                     db.close()
-                    os.remove(self.__dbFile)
+                    exc = sys.exc_info()[1]
+                    try:
+                        os.remove(self.__dbFile)
+                    except OSError:
+                        pass
                     raise Exception(
                         'Data error at %s:%d: %s' % (
-                            self.__textFile, lineNo, sys.exc_info()[1]
+                            self.__textFile, lineNo, exc
                             )
                         )
 
@@ -361,10 +366,14 @@ class DeviceFile:
                         textParser.evaluateOid(oid)
                     except Exception:
                         db.close()
-                        os.remove(self.__dbFile)
+                        exc = sys.exc_info()[1]
+                        try:
+                            os.remove(self.__dbFile)
+                        except OSError:
+                            pass
                         raise Exception(
                             'OID error at %s:%d: %s' % (
-                                self.__textFile, lineNo, sys.exc_info()[1]
+                                self.__textFile, lineNo, exc
                                 )
                             )
                     try:
@@ -442,9 +451,11 @@ def getDevices(tgtDir, topLen=None):
 # Lightweignt MIB instrumentation (API-compatible with pysnmp's)
 
 class MibInstrumController:
+    eol = str2octs('\n')
     def __init__(self, deviceFile, textParser):
         self.__deviceFile = deviceFile
         self.__textParser = textParser
+        self.__writeCache = {}
 
     def __str__(self): return str(self.__deviceFile)
 
@@ -459,7 +470,7 @@ class MibInstrumController:
             text.seek(mid)
             while mid:
                 c = text.read(1)
-                if c == '\n':
+                if c == self.eol:
                     mid = mid + 1
                     break
                 mid = mid - 1    # pivot stepping back in search for full line
@@ -527,13 +538,22 @@ class MibInstrumController:
                 continue
 
             try:
-                oid, val = self.__textParser.evaluate(line)
+                _oid, _val = self.__textParser.evaluate(line)
             except Exception:
                 raise Exception(
                     'Data error at %s for %s: %s' % (self, textOid, sys.exc_info()[1])
                     )
 
-            rspVarBinds.append((oid, val))
+            if writeMode:
+                try:
+                    self.__writeCache[_oid] = _val = _val.clone(val)
+                except PyAsn1Error:
+                    _val = errorStatus
+            else:
+                if _oid in self.__writeCache:
+                    _val = self.__writeCache[_oid]
+
+            rspVarBinds.append((_oid, _val))
 
         return rspVarBinds
     
@@ -581,6 +601,82 @@ class DevicesIndexInstrumController:
 
 devicesIndexInstrumController = DevicesIndexInstrumController()
 
+# Suggest variations of context name based on request data
+def probeContext(transportDomain, transportAddress, contextName):
+    candidate = [
+        contextName, '.'.join([ str(x) for x in transportDomain ])
+    ]
+    if transportDomain[:len(udp.domainName)] == udp.domainName:
+        candidate.append(transportAddress[0])
+    elif udp6 and transportDomain[:len(udp6.domainName)] == udp6.domainName:
+        candidate.append(
+            str(transportAddress[0]).replace(':', '_')
+        )
+    elif unix and transportDomain[:len(unix.domainName)] == unix.domainName:
+        candidate.append(transportAddress)
+
+    candidate = [ str(x) for x in candidate if x ]
+
+    while candidate:
+        yield rfc1902.OctetString(
+                  os.path.normpath(os.path.sep.join(candidate))
+              ).asOctets()
+        del candidate[-1]
+ 
+if not v2cArch:
+    def probeHashContext(self, snmpEngine, stateReference, contextName):
+        transportDomain, transportAddress = snmpEngine.msgAndPduDsp.getTransportInfo(stateReference)
+
+        for probedContextName in probeContext(transportDomain, transportAddress, contextName):
+            probedContextName = md5(probedContextName).hexdigest()
+            try:
+                self.snmpContext.getMibInstrum(probedContextName)
+            except error.PySnmpError:
+                pass
+            else:
+                return probedContextName
+        return contextName
+
+    class GetCommandResponder(cmdrsp.GetCommandResponder):
+        def handleMgmtOperation(
+                self, snmpEngine, stateReference, contextName, PDU, acInfo
+            ):
+            cmdrsp.GetCommandResponder.handleMgmtOperation(
+                self, snmpEngine, stateReference, 
+                probeHashContext(self, snmpEngine, stateReference, contextName),
+                PDU, acInfo
+            )
+
+    class SetCommandResponder(cmdrsp.SetCommandResponder):
+        def handleMgmtOperation(
+                self, snmpEngine, stateReference, contextName, PDU, acInfo
+            ):
+            cmdrsp.SetCommandResponder.handleMgmtOperation(
+                self, snmpEngine, stateReference, 
+                probeHashContext(self, snmpEngine, stateReference, contextName),
+                PDU, acInfo
+            )
+
+    class NextCommandResponder(cmdrsp.NextCommandResponder):
+        def handleMgmtOperation(
+                self, snmpEngine, stateReference, contextName, PDU, acInfo
+            ):
+            cmdrsp.NextCommandResponder.handleMgmtOperation(
+                self, snmpEngine, stateReference, 
+                probeHashContext(self, snmpEngine, stateReference, contextName),
+                PDU, acInfo
+            )
+
+    class BulkCommandResponder(cmdrsp.BulkCommandResponder):
+        def handleMgmtOperation(
+                self, snmpEngine, stateReference, contextName, PDU, acInfo
+            ):
+            cmdrsp.BulkCommandResponder.handleMgmtOperation(
+                self, snmpEngine, stateReference, 
+                probeHashContext(self, snmpEngine, stateReference, contextName),
+                PDU, acInfo
+            )
+
 # Basic SNMP engine configuration
 
 if v2cArch:
@@ -658,29 +754,6 @@ if v2cArch:
 
         return rspVarBinds
  
-    # Suggest variations of context name based on request data
-
-    def probeContext(transportDomain, transportAddress, community):
-        candidate = [
-            community, '.'.join([ str(x) for x in transportDomain ])
-        ]
-        if transportDomain[:len(udp.domainName)] == udp.domainName:
-            candidate.append(transportAddress[0])
-        elif udp6 and transportDomain[:len(udp6.domainName)] == udp6.domainName:
-            candidate.append(
-                str(transportAddress[0]).replace(':', '_')
-            )
-        elif unix and transportDomain[:len(unix.domainName)] == unix.domainName:
-            candidate.append(transportAddress)
-
-        candidate = [ str(x) for x in candidate ]
-
-        while candidate:
-            yield rfc1902.OctetString(
-                      os.path.normpath(os.path.sep.join(candidate))
-                  )
-            del candidate[-1]
-
     def commandResponderCbFun(transportDispatcher, transportDomain,
                               transportAddress, wholeMsg):
         while wholeMsg:
@@ -828,10 +901,10 @@ else:
 
     # SNMP applications
 
-    cmdrsp.GetCommandResponder(snmpEngine, snmpContext)
-    cmdrsp.SetCommandResponder(snmpEngine, snmpContext)
-    cmdrsp.NextCommandResponder(snmpEngine, snmpContext)
-    cmdrsp.BulkCommandResponder(snmpEngine, snmpContext)
+    GetCommandResponder(snmpEngine, snmpContext)
+    SetCommandResponder(snmpEngine, snmpContext)
+    NextCommandResponder(snmpEngine, snmpContext)
+    BulkCommandResponder(snmpEngine, snmpContext)
 
     transportDispatcher = snmpEngine.transportDispatcher
 
